#include <errno.h>
#include <error.h>
#include <inttypes.h>
#include <pthread.h>
#include <sched.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "tracer.h"

#define CLOCK_ID CLOCK_MONOTONIC
#define NSEC_PER_SEC INT64_C(1000000000)

typedef struct thread_stat {
  int nb_cycles;
  uint64_t max_res;
  uint64_t min_res;
} thread_stat_t;

typedef struct thread_param {
  int priority;
  int max_cycles;
  uint64_t interval;
  thread_stat_t stat;
} thread_param_t;

typedef struct main_param {
  int refresh_rate;
} main_param_t;

static void process_options(int argc, char *argv[], thread_param_t *param,
                            main_param_t *main_param);

static inline uint64_t calcdiff_ns(struct timespec t1, struct timespec t2);
static inline uint64_t max(uint64_t a, uint64_t b);
static inline uint64_t min(uint64_t a, uint64_t b);

// Real-time thread
static void *timerthread(void *p) {
  struct timespec previous, current;
  struct sched_param priority;
  uint64_t diff;

  thread_param_t *param = (thread_param_t *)p;
  thread_stat_t *stat = &param->stat;

  priority.sched_priority = param->priority;

  if (sched_setscheduler(0, SCHED_FIFO, &priority))
    error(EXIT_FAILURE, errno, "Couldn't set priority");

  stat->max_res = 0;
  stat->min_res = UINT64_MAX;

  for (stat->nb_cycles = 0;; stat->nb_cycles++) {
    if (param->max_cycles && (stat->nb_cycles >= param->max_cycles)) break;

    tracing(1);
    clock_gettime(CLOCK_ID, &previous);
    clock_gettime(CLOCK_ID, &current);
    tracing(0);

    diff = calcdiff_ns(current, previous);

    stat->max_res = max(stat->max_res, diff);
    stat->min_res = min(stat->min_res, diff);

    usleep(param->interval);
  }

  return NULL;
}

// Main thread, has non-real time priority
// Handles the IO and creates real time threads
int main(int argc, char *argv[]) {
  pthread_t thread;
  thread_param_t param;
  main_param_t main_param;

  // Default values
  param.interval = 1000;
  param.max_cycles = 100000;
  param.priority = 99;
  main_param.refresh_rate = 10000;

  process_options(argc, argv, &param, &main_param);
  setup_tracer();

  usleep(10000);

  if (pthread_create(&thread, NULL, timerthread, (void *)&param))
    error(EXIT_FAILURE, errno, "Couldn't create thread");

  for (;;) {
    usleep(main_param.refresh_rate);

    printf("Maximum res: %" PRIu64 "ns (%d)", (param.stat.max_res),
           param.stat.nb_cycles);
    printf(", minimum res: %" PRIu64 "ns (%d)\n", (param.stat.min_res),
           param.stat.nb_cycles);

    if (param.max_cycles == param.stat.nb_cycles) break;
  }

  tracing(0);

  exit(EXIT_SUCCESS);
}

static void process_options(int argc, char *argv[], thread_param_t *param,
                            main_param_t *main_param) {
  for (;;) {
    int c = getopt(argc, argv, "l:p:i:r:");

    if (c == -1) break;

    switch (c) {
      case 'p':
        param->priority = atoi(optarg);
        break;
      case 'l':
        param->max_cycles = atoi(optarg);
        break;
      case 'i':
        param->interval = atoi(optarg);
        break;
      case 'r':
        main_param->refresh_rate = atoi(optarg);
        break;
      default:
        exit(EXIT_FAILURE);
        break;
    }
  }
}

static inline uint64_t calcdiff_ns(struct timespec t1, struct timespec t2) {
  uint64_t diff;
  diff = NSEC_PER_SEC * (uint64_t)((int)t1.tv_sec - (int)t2.tv_sec);
  diff += ((int)t1.tv_nsec - (int)t2.tv_nsec);
  return diff;
}

static inline uint64_t max(uint64_t a, uint64_t b) { return a > b ? a : b; }
static inline uint64_t min(uint64_t a, uint64_t b) { return a < b ? a : b; }

